#coding: utf-8

from report_generator.core.query_components import *

class EntityTypes(object):

    MODEL = 1
    QUERY = 2

    TYPES = {
        MODEL: u'Модели проекта',
        QUERY: u'Запросы'
    }

    # Сопоставление типа сущности и класса. Используется в фабрике сущностей.
    CLASS_MAP = {
        MODEL: TableEntity,
        QUERY: QueryEntity
    }

    @classmethod
    def get_entity_type_title(cls, type):
        """
        Возвращает заголовок для сущностей
        """

        return EntityTypes.TYPES.get(type)


class EntityFactory(object):
    """
    Фабрика потомков constructor.core.query_component.BaseAlchemyEntity
    """

    @classmethod
    def create_entity(cls, entity_id, entity_type):
        """
        По JSON представлению сущности строим потомка BaseAlchemyEntity
        """

        assert entity_type in EntityTypes.TYPES, u'Не поддерживаемый тип сущности'

        entity_class = EntityTypes.CLASS_MAP.get(entity_type)
        return entity_class(entity_id, alchemy_query_builder=AlchemyQueryBuilder)


class AlchemyQueryBuilder(object):
    """
    Превращает некий обьект в AlchemyQuery
    """

    def __init__(self, name, **kwargs):

        self.json_query = kwargs.get('json_query')
        self.alchemy_query = AlchemyQuery(name)

    def build(self):
        """
        """
        self._prepare_internal_entities()
        self._prepare_relations()
        self._prepare_group_by()
        self._prepare_select_and_order_fields()
        self._prepare_where()
        self._prepare_limit()
        self._prepare_distinct()

        return self.alchemy_query

    def _prepare_internal_entities(self):
        """
        Подготовка подсущностей
        """

        self.alchemy_query.entities = [self._prepare_entity_object(x) for x
            in self.json_query.get('entities')]

    def _prepare_entity_object(self, entity):
        """
        """

        entity_id = entity.get('entityId')
        entity_type = entity.get('entityType')

        return EntityFactory.create_entity(entity_id, entity_type)

    def _prepare_relations(self):
        """
        """

        relations = self.json_query.get('relations')

        def populate(relation):
            first_part, second_part = relation.get('id').split('|')
            first_entity_id, first_field_id = first_part.split('-')

            second_entity_id, second_field_id = second_part.split('-')

            first_outer = relation.get('outerFirst')
            second_outer = relation.get('outerSecond')

            self.alchemy_query.relations.append(
                Relation(
                    self._prepare_field(field_id=first_field_id, entity_type=relation.get('leftEntityType')),
                    self._prepare_field(field_id=second_field_id, entity_type=relation.get('rightEntityType')),
                    outer_first=first_outer,
                    outer_second=second_outer)
            )

        map(populate, relations)

    def _prepare_field(self, field_id, entity_type, alias=None):

        app, model, field = field_id.split('.')
        entity_id = '.'.join([app, model])

        entity = EntityFactory.create_entity(entity_id, entity_type)

        return Field(entity=entity, field_id=field_id, alias=alias)

    def _prepare_group_by(self):

        self.alchemy_query.group_by = Grouping(group_fields=self._prepare_group_fields(),
            aggregate_fields=self._prepare_aggregation_fields())

    def _prepare_group_fields(self):

        group_fields = []

        def populate(group_field):
            field = self._prepare_field(group_field.get('id'), group_field.get('entityType'))
            group_fields.append(field)

        map(populate, self.json_query.get('group').get('group_fields'))

        return group_fields

    def _prepare_aggregation_fields(self):

        aggregation_fields = []

        def populate(aggr_field):

            field = self._prepare_field(aggr_field.get('id'), entity_type=aggr_field.get('entityType'),
                alias=aggr_field.get('alias'))
            aggr_func = aggr_field.get('function')

            if aggr_func:
                aggr_class = Grouping.get_aggr_functions()[aggr_func]
                aggregation_fields.append(aggr_class(field))

        map(populate, self.json_query.get('group').get('group_aggr_fields'))

        return aggregation_fields

    def _prepare_select_and_order_fields(self):

       def populate(select_field):

           field = self._prepare_field(select_field.get('id'), entity_type=select_field.get('entityType'),
                alias=select_field.get('alias'))
           self.alchemy_query.select.append(field)

           sort_order = SortOrder.get_order_code(select_field.get('sorting'))
           if sort_order:
                self.alchemy_query.order_by.append(SortOrder(field, order=sort_order))

       map(populate, self.json_query.get('selected_fields'))

    def _prepare_limit(self):

        self.alchemy_query.limit = self.json_query.get('limit')

    def _prepare_distinct(self):

        self.alchemy_query.distinct = self.json_query.get('distinct')

    def _prepare_where(self):

        conditions = self.json_query.get('cond_fields')

        def populate(condition):

            if condition.get('param_type') == Where.VALUE_PARAM:
                right = Param(name=condition.get('parameter'), type=None, verbose_name=None)
            else:
                right = self._prepare_field(condition.get('parameter'),
                                                                 entity_type=condition.get('second_field_entity_type'))

            where = Where(left=self._prepare_field(condition.get('id_field'), condition.get('first_field_entity_type')),
                op=condition.get('condition'), right=right)

            if not self.alchemy_query.where:
                self.alchemy_query.where = where
            else:
                if condition.get('connector') == Where.AND:
                    self.alchemy_query.where &= where
                elif condition.get('connector') == Where.OR:
                    self.alchemy_query.where |= where

        map(populate, conditions)