#coding: utf-8

import json

from django.db.models import loading

from m3.helpers.icons import Icons

import settings

from report_generator.core.entity_proxy import EntityUIProxyInterface
from report_generator.core.builder import EntityTypes
from report_generator.query_builder.models import Query
from report_generator.query_builder.proxy import QueryProxy
from report_generator.constructor_settings import MODELS, EXCLUDE_MODELS

class QueryEntityUIProxy(EntityUIProxyInterface):
    """
    Посредник между запросами созданными в КО и UI

    Протокол хранения:
    1) Сущность-запрос
        а) Идентификатор - queryId.queryName - queryId - идентификатор записи в БД,
                                                         атрибут name записи в БД
        б) Название атрибут name записи в БД.
    2) Поле сущности-запроса
    """

    @classmethod
    def get_entities(cls):
        """
        Подготавливает данные о имеющихся запросах для загрузки в дерево сущностей.
        """

        # Получаем все запросы, созданные при помощи КО.
        queries = Query.objects.values('id', 'name')

        # Корневой узел
        root_node = dict(id = EntityTypes.QUERY,
                         iconCls = Icons.FOLDER_TABLE,
                         title = EntityTypes.get_entity_type_title(EntityTypes.QUERY),
                         leaf = False,
                         expanded = True)

        def populate(query):
            query_name = query.get('name')
            query_id = query.get('id')
            node = dict(id='.'.join([str(query_id), query_name]),
                        leaf=True,
                        iconCls=Icons.PLUGIN,
                        title=query_name)

            root_node.setdefault('children', []).append(node)

        # Заполняем запросы
        map(populate, queries)

        # Если запросов , не имеется, то корневой узел добавлять не будем
        if not root_node.get('children'):
            root_node = None

        return root_node

    @classmethod
    def get_entity_fields(cls, query):
        """
        Формируем данные о полях запроса
        @param: query запрос, созданный КО, в формате constructor.query_builder.model.Query
        """

        assert isinstance(query, Query)

        # Корневой узел. Содержит имя запроса
        root_node = dict(id='.'.join([str(query.id), query.name]),
                         leaf=False,
                         iconCls=Icons.PLUGIN,
                         verbose_field=query.name,
                         expanded=True)

        def populate(field):
            """
            """

            id, verbose_field, alias, _, entity_type, entity_id = field
            entity_id = '.'.join([str(query.id), query.name])

            node = dict(id_field = '.'.join([entity_id, alias or verbose_field]),
                        leaf=True,
                        verbose_field=alias or verbose_field, entity_type=EntityTypes.QUERY,
                        entity_id = entity_id)

            root_node.setdefault('children', []).append(node)

        # Используя прокси класс получаем доступ к select полям.
        # Select-поля всегда имеются. Иначе клиентская часть сообщит о некорректности запроса при сохранении
        fields = QueryProxy(query_id=query.id, query_name=query.name,
                                                        json_query=json.loads(query.json_query)).get_selected_fields()

        # Заполняем поля запроса
        map(populate, fields)

        # Возвращаем корневой узел
        return root_node

    @classmethod
    def get_entities_fields(cls, ids):
        """
        Для выбранных сущностей, готовим список их select полей
        """

        # Идентификатор сущности-запроса имеет вид id.name
        ids = [int(id.split('.')[0]) for id in ids]

        # Получаем список необходимых нам запросов
        queries = Query.objects.filter(id__in=ids)

        return [cls.get_entity_fields(query) for query in queries]


class ModelEntityUIProxy(EntityUIProxyInterface):
    """
    Посредник между моделями проекта и UI

    Протокол хранения:
    1)Сущность-модель
        a) Идентификатор  id    = имяПриложения.имяМодели
        б) Название       title = имяМодели
    2)Поле сущности-модели
        a) Идентификатор               id = имяПоля.идентификаторМодели
        б) Идентификатор модели entity_id = идентификаторМодели
        в) Название поля     verbose_field = имяПоля
    """

    # Модели, которые не нужно добавлять. Внутренние модели конструктора.
    EXCLUDE_MODEL_LIST = [
        'Query',
        'Report',
        'ReportQuery',
        'Template',
        'FormField',
        'SubstValues'
    ]

    @classmethod
    def get_entities(cls):
        """
        Получаем список всех моделей
        """

        root_node = dict(id = EntityTypes.MODEL,
            iconCls = Icons.FOLDER_TABLE,
            title = EntityTypes.get_entity_type_title(EntityTypes.MODEL),
            leaf = False,
            expanded = True)

        def populate(model):

            node = dict(id = '.'.join([model._meta.app_label, model.__name__]),
                leaf = True,
                iconCls = Icons.PLUGIN,
                title = model.__name__)

            root_node.setdefault('children', []).append(node)

        map(populate, cls._filter_models())

        if not root_node.get('children'):
            root_node = None

        return root_node

    @classmethod
    def _filter_models(cls):
        """
        Фильтр моделей
        """

        # Все модели проекта и окружения
        models = loading.get_models()

        for model in models:
            model_name = model.__name__
            # Добавляем модели не принадлежащие конструктору и не указанные в списке исключений в настройках
            # прикладного проекта.
            if model_name not in cls.EXCLUDE_MODEL_LIST and model_name not in EXCLUDE_MODELS:
                # Если в настройках прикладного проекта указан список необходимых моделей, то фильтруем по
                # этому списку. В противном случае берём все.
                if MODELS and model_name in MODELS or not MODELS:
                    yield model

    @classmethod
    def get_entity_fields(cls, model_id):
        """
        Принимает имена модели и приложения, возвращает список полей для UI
        """

        app_name, model_name = model_id.split('.')

        model_class = loading.get_model(app_name, model_name)

        root_node = dict(id = model_id,
            leaf = False,
            iconCls = Icons.PLUGIN,
            verbose_field = model_name,
            expanded = True)

        def populate(field):

            node = dict(id_field = '.'.join([model_id, field.attname]),
                verbose_field = getattr(field, 'verbose_name') or field.name,
                leaf = True, entity_type = EntityTypes.MODEL,
                entity_id = model_id)

            root_node.setdefault('children', []).append(node)

        map(populate, model_class._meta.fields)

        return root_node

    @classmethod
    def get_entities_fields(cls, models_ids):
        """
        """

        return [cls.get_entity_fields(model) for model in models_ids]